:- use_module(library(clpfd)).
:- use_module(library(lists)).

disj_spec(X, Y) :-
	domain([X,Y], 0, 6),
	(   X+5 #=< Y
	;   Y+5 #=< X
	).

disj_reif(X, Y) :-
	domain([X,Y], 0, 6),
	X+5 #=< Y #\/ Y+5 #=< X.

disj_abs(X, Y) :-
	domain([X,Y], 0, 6),
	scalar_product([1,-1],[X,Y],#=,D,[consistency(domain)]), 
        abs(D) #>= 5.

% ---------------------------------------------------------------------------

% Perform constructive disjunction of constraints Cs
% with respect to variable Var.
cdisj(Cs, Var) :-
    findall(S, (member(C,Cs),C,fd_set(Var,S)), Doms),
    fdset_union(Doms,Set),
    Var in_set Set.

% ?- domain([X,Y],0,6), cdisj([X+5#=<Y,Y+5#=<X], X).

% Shave value V off the variable X.
shave_value(X, V) :-
    (   \+ (X = V) -> X #\= V
    ;   true
    ).
	
% Shave off variable X all values within the domain of X.
% Equivalent to cdisj([X=V1,...X=Vn], X), where the domain
% of X is {V1,...,Vn}.
shave_all_values0(X) :-
	fd_set(X, FD),
	fdset_to_list(FD, L),
	maplist(shave_value(X), L).    % i.e., if L = [A,B,C,...] this is equivalent to:
	                               % shave_value(X, A),
	                               % shave_value(X, B),
	                               % shave_value(X, C),
	                               % ...

% Shave all values for variable X.
% Equivalent to shave_all_values0/1, but probably more efficient.
shave_all_values(X) :-
	fd_set(X, FD),
	findall(X, fdset_member(X,FD), Vs),
	list_to_fdset(Vs, FD1),
	X in_set FD1.

% Shave all values for variable X.
% Equivalent to shave_all_values/1, and works in both SWI and SICStus Prolog systems.
% (In SICStus, slower than shave_all_values/1.)
shave_all_values_swi(X) :-
        findall(X, indomain(X), ShavedList),
        ShavedList \== [],
        list_to_round_list(ShavedList, Enumeration),
        X in {Enumeration}.

% list_to_round_list(+L, ?RL): RL is the "round list" having the same
% elements in the same order as proper list L. Example (run in SICStus):
%      | ?- list_to_round_list([a,b,c], RL), write_canonical(RL).
%      ','(a,','(b,c))
%      RL = (a,b,c) ? 
list_to_round_list([C], C) :- !.
list_to_round_list([C|T], (C,RT)) :-
        list_to_round_list(T, RT).



% Perform shave_all_values(X), for all elements X of list L.
shave_all_elements(L) :-
	maplist(shave_all_values, L).


% ---------------------------------------------------------------------------
% Bent triples (Y wing) sudoku technique -- example

sudoku_initial_state([R13,R15,R16,R26,R33,R36]) :-
    R13 in {6,7}, R15 in {1,2,6}, R16 in {2,3,6},
    R26 in {4,5,6}, R33 in {7,8}, R36 in {6,8},
    all_distinct([R13,R15,R16]),          % row 1
    all_distinct([R33,R36]),              % row 3
    all_distinct([R13,R33]),              % box 1
    all_distinct([R15,R16,R26,R36]).      % box 2

sudoku_example :-
    VarNames = ['R13','R15','R16','R26','R33','R36'],
    sudoku_initial_state(St),
    maplist(fd_dom, St,  InitSt),
    shave_all_elements(St),
    maplist(fd_dom, St,  ShavedSt),
    transpose([VarNames,InitSt,ShavedSt], Prunings),
    (   nth1(N, [['VAR','Before shave','After shave']|Prunings], [VName,DomI,DomSh]),
        format('~w ~| ~w~t~14+ ~w~t~14+\n',[VName,DomI,DomSh]), 
        N = 1, write('------------------------------\n'),
        fail
    ;   true
    ).

    
% ---------------------------------------------------------------------------

% L is a list of N distinct digits 1..9 with a sum Sum.
kakuro(N, L, Sum) :-
	length(L, N),
	domain(L, 1, 9),
	all_distinct(L),
	sum(L, #=, Sum).

% Sample constraints for a kakuro "word" of length 4
sample_domains(L) :-
	L = [A,B,C,D],
	A in {5,9},
	B in 1..9,
	C in {6,8,9},
	D = 4.

% Test how much domain pruning is obtained for the sample.
test1(L) :-
	kakuro(4, L, 23),
	sample_domains(L).

% Test how much domain pruning is obtained for variable B, 
% if a constructive disjunction A = 5 \/ A = 9 is applied
test2(L) :-
	test1(L),
	L = [A,B|_],
	cdisj([A=5,A=9], B).


% Test how much domain pruning is obtained
% when value 9 is shaved off variable C.
test3(L) :-
	test1(L),
	L = [_,_,C,_],
	shave_value(C, 9).

% Test how much domain pruning is obtained when
% all values of all variables are shaved.
test4(L) :-
	test1(L),
	shave_all_elements(L).

% kakuro_table(+N, +Sum, -Table):
% Table is a list of all n-element lists of digits 1..9
% with distinct elements and a sum Sum.
kakuro_table(N, Sum, Table) :-
	findall(L, (   kakuro(N, L, Sum),
		       labeling([], L)
		   ),
		Table).

% A domain consistent version of kakuro(N, L, Sum).
kakuro_dc(N, L, Sum) :-
	length(L, N),
	kakuro_table(N, Sum, Table),
	table([L], Table).

% Test how much domain pruning is obtained
% with the domain consistent kakuro constraint.
test5(L) :-
	kakuro_dc(4, L, 23),
	sample_domains(L).


% Test how much domain pruning is obtained for variable X
% if a constructive disjunction 'X+5#=<Y or Y+5#=<X' is applied.
test6(X, Y) :-
	L = [X,Y],
	domain(L, 0, 6),
	cdisj([X+5#=<Y,Y+5#=<X], X).  


end_of_file.

| ?- disj_spec(X, Y).
X in 0..1,
Y in 5..6 ? ;
X in 5..6,
Y in 0..1 ? ;
no
| ?- disj_reif(X, Y).
X in 0..6,
Y in 0..6 ? ;
no
| ?- disj_abs(X, Y).
X in(0..1)\/(5..6),
Y in(0..1)\/(5..6) ? ;
no
| ?- test1(L).
L = [_A,_B,_C,4],
_A in{5}\/{9},
_B in(1..3)\/(5..8),
_C in{6}\/(8..9) ? ;
no
| ?- test2(L).
L = [_A,_B,_C,4],
_A in{5}\/{9},
_B in{2}\/(6..8),
_C in{6}\/(8..9) ? ;
no
| ?- test3(L).
L = [_A,_B,_C,4],
_A in{5}\/{9},
_B in(2..3)\/(5..8),
_C in{6}\/{8} ? ;
no
| ?- test4(L).
L = [_A,_B,_C,4],
_A in{5}\/{9},
_B in{2}\/{6}\/{8},
_C in{6}\/{8} ? ;
no
| ?- kakuro(4, L, 23), sample_domains(L), labeling([], L).
L = [5,6,8,4] ? ;
L = [5,8,6,4] ? ;
L = [9,2,8,4] ? ;
no
| ?- test5(L).
L = [_A,_B,_C,4],
_A in{5}\/{9},
_B in{2}\/{6}\/{8},
_C in{6}\/{8} ? ;
no
| ?- test6(X, Y).
X in(0..1)\/(5..6),
Y in 0..6 ? ;
no
